/*
* dvb-aspect.c:  Tool to grab the aspect ratio of video played on a dvb ff (av7110)
*
* Copyright (C) 2006 Jens Vogel (jens.vogel@justmail.de)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
* Or, point your browser to http://www.gnu.org/copyleft/gpl.html
*/

#include <stdio.h>
#include <float.h>
#include <math.h>
#include <memory.h>
#include <getopt.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/dvb/video.h>
#include <linux/videodev.h>

#define DEFAULT_VIDEO_DEVICE "/dev/video0"
#define DEFAULT_DVB_DEVICE  "/dev/dvb/adapter0/video0"
#define DEFAULT_BLACK_VIDEO 0x15

typedef struct 
{
  float       fMinAspect;
  const char* pszName;
} SAspectInfo;
SAspectInfo aAspectInfo[] =
{
  { 1.46,     "4:3" },
  { 1.66,     "L14:9" },
  { 1.90,     "L16:9" },
  { FLT_MAX,  "L>16:9" }
};
const int naAspectInfo = sizeof(aAspectInfo) / sizeof(SAspectInfo);

int main(int argc, char **argv)
{
  char *pszDevVideo = "/dev/video0";
  char *pszDevDVB = "/dev/dvb/adapter0/video0";
  unsigned char blackVideo = 0x15;
  int bNativeAspectOutput = 0;

  int bMoreOptions = true;
  opterr = 0;
  do
  {
    static struct option aOptions[] =
    {
      { "dev_video", required_argument, 0, 'v' },
      { "dev_dvb", required_argument, 0, 'd' },
      { "black", required_argument, 0, 'b' },
      { "aspect", no_argument, 0, 'a' },
      { "help", no_argument, 0, 'h' },
      { NULL }
    };
    switch(getopt_long(argc, argv, "v:d:b:ah", aOptions, NULL))
    {
    case 'v':
      pszDevVideo = optarg;
      break;
    case 'd':
      pszDevDVB = optarg;
      break;
    case 'b':
      sscanf(optarg, "%x", &blackVideo);
      break;
    case 'a':
      bNativeAspectOutput = 1;
      break;
    case 'h':
      printf("Usage: %s [<Options>]\n"
             "  -v, --dev_video=<Device>  Video device\n"
             "                            (Default is " DEFAULT_VIDEO_DEVICE ")\n" 
             "  -d, --dev_dvb=<Device>    DVB device\n"
             "                            (Default is " DEFAULT_DVB_DEVICE ")\n"
             "  -b, --black=<Value>       Lowest value of a pixel, interpreted as black\n"
             "                            (hex; grayscale 0..ff; Default is %x)\n"
             "  -a, --aspect              Output aspect ratio in addition as number\n"
             "                            (e.g \"133 4:3\")\n"
             "  -h, --help                Displays this page (and exits)\n",
             argv[0],
             DEFAULT_BLACK_VIDEO);
      return 0;
      break;
    case '?':
      fprintf(stderr, "%s: error - unkown option %s [arg #%d] - ignored\n", argv[0], argv[optind-1], optind-1);
      break;
    case ':':
      fprintf(stderr, "%s: error - missing argument for option %s [arg #%d] - ignored\n", argv[0], argv[optind-1], optind-1);
      break;
    default:
      bMoreOptions = false;
      break;
    }
  }
  while(bMoreOptions);
  if (optind < argc)
    fprintf(stderr, "%s: error - extra arguments '%s'... [arg #%d...] - ignored\n", argv[0], argv[optind], optind);


  int devDVB = open(pszDevDVB,O_RDONLY|O_NONBLOCK);
  if (devDVB < 0)
    fprintf(stderr, "Error: Can't open dvb device '%s' (%s)\n", pszDevDVB, strerror(errno));
  else
  {
    int devVideo = open(pszDevVideo, O_RDWR);
    if (devVideo < 0)
      fprintf(stderr, "Error: Can't open video device '%s' (%s)\n", pszDevVideo, strerror(errno));
    else 
    {
      struct video_mbuf mbufVideo;
      if (ioctl(devVideo, VIDIOCGMBUF, &mbufVideo) != 0) 
        fprintf(stderr, "Error: Can't open video buffer on video device '%s' (%s)\n", pszDevVideo, strerror(errno));
      else
      {
        unsigned char *pbVideo = (unsigned char *)mmap(0, mbufVideo.size, PROT_READ | PROT_WRITE, MAP_SHARED, devVideo, 0);
        if (!pbVideo || (pbVideo == (unsigned char *)-1))
          fprintf(stderr, "Error: Can't map video buffer of video device '%s' (%s)\n", pszDevVideo, strerror(errno));
        else
        {
          struct video_capability capabilityVideo;
          if (ioctl(devVideo, VIDIOCGCAP, &capabilityVideo))
            fprintf(stderr, "Error: Can't query caps of video device '%s' (%s)\n", pszDevVideo, strerror(errno));
          else
          {
            struct video_mmap mmapVideo;
            mmapVideo.frame = 0;
            mmapVideo.width = capabilityVideo.maxwidth;
            mmapVideo.height = capabilityVideo.maxheight;
            mmapVideo.format = VIDEO_PALETTE_GREY;
            int sizeVideo = mmapVideo.width * mmapVideo.height;
            int bAnamorph = 0;
            int iAspectInfo = -1;
            float fNativeAspectOutput = 0;
            int bError = 0;
            while (!bError)
            {
              int bFormatChanged = 0;
              struct video_status statusVideo;
              if (bError = (ioctl(devDVB, VIDEO_GET_STATUS, &statusVideo) < 0)) 
                fprintf(stderr, "Error: Can't get status of device '%s' (%s)\n", pszDevDVB, strerror(errno));
              else
              {
                switch (statusVideo.video_format)
                {
                case VIDEO_FORMAT_16_9:
                case VIDEO_FORMAT_221_1:
                  if (!bAnamorph)
                  {
                    bFormatChanged = 1;
                    bAnamorph = 1;
                  }
                  break;
                case 	VIDEO_FORMAT_4_3:
                default:
                  if (bAnamorph)
                  {
                    iAspectInfo = -1;
                    bAnamorph = 0;
                  }
                  if (bError = (  ioctl(devVideo, VIDIOCMCAPTURE, &mmapVideo)
                                | ioctl(devVideo, VIDIOCSYNC, &mmapVideo.frame)))
                    fprintf(stderr, "Error: Can't grab on video device '%s' (%s)\n", pszDevVideo, strerror(errno));
                  else
                  {
                    int iBlackTop = 0; 
                    while ((iBlackTop < sizeVideo/2) && (pbVideo[iBlackTop] <= blackVideo))
                      iBlackTop++;
                    int iBlackBottom = sizeVideo-1;
                    while ((iBlackBottom > sizeVideo/2) && (pbVideo[iBlackBottom] <= blackVideo))
                      iBlackBottom--;
                    iBlackBottom = sizeVideo - iBlackBottom;
                    int iBlackLinesTop = iBlackTop/mmapVideo.width;
                    int iBlackLinesBottom = iBlackBottom/mmapVideo.width;
                    int iBlackLines;
                    if (iBlackLinesTop > iBlackLinesBottom)
                      iBlackLines = 2*iBlackLinesBottom;
                    else
                      iBlackLines = 2*iBlackLinesTop;
                    int nActiveLines = mmapVideo.height - iBlackLines;
                    if (nActiveLines > 0)
                    {
                      const float fLinearWidth = 4./3.*mmapVideo.height;
                      if (iAspectInfo < 0)
                        bFormatChanged = 1;
                      else
                      {
                        const int nAspectHyst = 2;
                        int nMinActiveLines = (int)(fLinearWidth/aAspectInfo[iAspectInfo].fMinAspect) - nAspectHyst;
                        int nMaxActiveLines =   (iAspectInfo > 0)
                                              ? (int)(fLinearWidth/aAspectInfo[iAspectInfo-1].fMinAspect) + nAspectHyst
                                              : mmapVideo.height;
                        bFormatChanged = (nActiveLines < nMinActiveLines) || (nActiveLines > nMaxActiveLines); 
                      }
                      float fAspect = fLinearWidth/nActiveLines;
                      if (bNativeAspectOutput)
                      {
                        float fNewNativeAspectOutput = floor(fAspect*100); 
                        if (!bFormatChanged)
                          bFormatChanged = fNewNativeAspectOutput != fNativeAspectOutput;
                        fNativeAspectOutput = fNewNativeAspectOutput;
                      }
                      if (bFormatChanged)
                      {
                        iAspectInfo = 0;
                        while (fAspect > aAspectInfo[iAspectInfo].fMinAspect)
                          iAspectInfo++;
                      }
                    }
                  }
                  break;
                }
              }
              if (!bError)
              {
                if (bFormatChanged)
                {
                  if (bAnamorph)
                    printf("16:9\n");
                  else
                  {
                    const char* pszAspect = aAspectInfo[iAspectInfo].pszName;
                    if (bNativeAspectOutput)
                      printf("%.0f %s\n", fNativeAspectOutput, pszAspect);
                    else
                      printf("%s\n", pszAspect);
                  }
                  fflush(stdout);
                }
                usleep(5000000);
              }
            }
            munmap(pbVideo, mbufVideo.size);
          }
        }
      }
      close(devVideo);
    }
    close(devDVB);
  }
  return 0;
}
